/*
Copyright (c) 1997-2005, Damon Hart-Davis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.hd.d.tmf;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.net.InetAddress;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: DHD
 * Date: 10-Jul-2005
 * Time: 15:32:28
 */

/**Filter to monitor throughput (output rate) to help regulate flow.
 * Can be used stand-alone,
 * or can be derived from to support more complex behaviour.
 * <p>
 * Can be used in association with an adaptive compression filter to
 * enable compression when our output pipe is likely to get congested.
 * <p>
 * This also makes a rough estimate of the number of users
 * (without forcing a session) by counting the number of IP addresses
 * used to visit the site through <em>this servlet container instance</em>
 * (probably a slight overestimate).  This ignores clients that are
 * probaby spiders (ie that never send a referring URL),
 * and that appear to be one-off hotlinks from external sites.
 * <p>
 * A truely distributed front end will need a more sophisticated communication
 * mechanism to share data efficiently between front-end instances.
 * <p>
 * This implementation essentially assumes that once loaded the
 * filter stays loaded unless there really is no activity
 * for a long time.
 * <p>
 * This filter can be supplied with a number of <code>init-param</code>
 * values:
 * <ul>
 * <li><b>maxBandwidth</b>:
 *     Available bandwidth in bytes per second for
 *     outgoing traffic, including headers and
 *     other HTTP/TCP overhead.
 *     This can be overridden by a system-property on
 *     a per-JVM basis.
 *     </li>
 * <li><b>maxBandwidthPropertyName</b>
 *     If this system property value is defined and
 *     usable then it overrides the maxBandwidth value.
 *     </li>
 * <li><b>upperThreshold</b>
 *     A value in the open range )0.0, 1.0(
 *     as a fraction of the overload traffic level
 *     above which traffic has to rise before
 *     the overload flag is set true.  Typical values might
 *     be .5 to .95; this will be clamped to
 *     a reasonable value internally.
 *     Must be higher than upperThreshold.
 *     </li>
 * <li><b>lowerThreshold</b>
 *     A value in the open range )0.0, 1.0(
 *     as a fraction of the overload traffic level
 *     below which traffic has to fall before
 *     the overload flag is set false.  Typical values might
 *     be .3 to .7; this will be clamped to
 *     a reasonable value internally.
 *     Must be lower than upperThreshold.
 *     </li>
 * <li><b>overloadFlagName</b>
 *     Name of attribute at application scope to set
 *     to Boolean.TRUE if we exceed our maxBandwidth
 *     and to set to Boolean.FALSE if we drop below our
 *     maxBandwidth * lowerThreshold.
 *     </li>
 * <li><b>throughputAttrNameBase</b>
 *     This creates several attributes at application scope
 *     with the given name plus a suffix,
 *     each updated about once per second.
 *     <dl>
 *     <dt>[name].Bps<dd>Integer: gives the current bandwidth in bytes/second.
 *     <dt>[name].ltBps<dd>Float: gives the long-term mean bandwidth in bytes/second.
 *     <dt>[name].totalBpsOutbound<dd>Float: gives the server-lifetime mean bandwidth in bytes/second.
 *     <dt>[name].connLoad  <dd>Float: gives the mean concurrent connection load.
 *     <dt>[name].ltConnLoad<dd>Float: gives the long-term mean concurrent connection load.
 *     <dt>[name].nonIdleTicks<dd>Float: gives the "busy-ness" of the system; 0.0 is idle, 1.0 is busy.
 *     <dt>[name].dailyUniqueVisitors<dd>Integer: gives the estimated number of unique users per day.
 *     </dl>
 * <li><b>clientIPPurgeTime</b>
 *     Time in seconds before an unused client IP address is purged
 *     from our records, ie before we regard the visitor has having left.
 *     The default is of the order of 120.
 * </ul>
 * <p>
 * For an example of use see the
 * <a href="http://gallery.hd.org/">DHD Multimedia Gallery</a>
 * and in particular the
 * <a href="http://gallery.hd.org/_javadoc/org/hd/d/pg2k/webSvr/util/ThroughputMonitorFilterPG2K.html">ThroughputMonitorFilterPG2K</a> class
 * which extends this base class,
 * and how it used in the <code>web.xml</code> configuration file like so:
<pre>

 &lt;filter&gt;
     &lt;filter-name&gt;Throughput Monitor Filter&lt;/filter-name&gt;
     &lt;filter-class&gt;org.hd.d.pg2k.webSvr.util.ThroughputMonitorFilterPG2K&lt;/filter-class&gt;
     &lt;init-param&gt;
         &lt;param-name&gt;maxBandwidth&lt;/param-name&gt;
         &lt;param-value&gt;6000&lt;/param-value&gt;
         &lt;description&gt;
             Available bandwidth in bytes per second for
             outgoing traffic, including an estimate for
             headers and other HTTP/TCP overhead.
             This can be overridden by a system-property on
             a per-JVM basis.
         &lt;/description&gt;
     &lt;/init-param&gt;
     &lt;init-param&gt;
         &lt;param-name&gt;maxBandwidthPropertyName&lt;/param-name&gt;
         &lt;param-value&gt;org.hd.d.pg2k.maxBytesPerSec&lt;/param-value&gt;
         &lt;description&gt;
             If this system property value is defined and
             usable it overrides the maxBandwidth value.
         &lt;/description&gt;
     &lt;/init-param&gt;
     &lt;init-param&gt;
         &lt;param-name&gt;upperThreshold&lt;/param-name&gt;
         &lt;param-value&gt;.95&lt;/param-value&gt;
         &lt;description&gt;
             A value in the open range )0.0, 1.0(
             as a fraction of the overload traffic level
             above which traffic has to rise before
             the overload flag is set true.  Typical values might
             be .5 to .95; this will be clamped to
             a reasonable value internally.
             Must be higher than lowerThreshold.
         &lt;/description&gt;
     &lt;/init-param&gt;
     &lt;init-param&gt;
         &lt;param-name&gt;lowerThreshold&lt;/param-name&gt;
         &lt;param-value&gt;.75&lt;/param-value&gt;
         &lt;description&gt;
             A value in the open range )0.0, 1.0(
             as a fraction of the overload traffic level
             below which traffic has to fall before
             the overload flag is set false.  Typical values might
             be .3 to .7; this will be clamped to
             a reasonable value internally.
             Must be lower than upperThreshold.
         &lt;/description&gt;
     &lt;/init-param&gt;
     &lt;init-param&gt;
         &lt;param-name&gt;overloadFlagName&lt;/param-name&gt;
         &lt;param-value&gt;org.hd.d.pg2k.highTraffic&lt;/param-value&gt;
         &lt;description&gt;
             Name of attribute at application scope to set
             to Boolean.TRUE if we exceed our maxBandwidth
             (and/or we exceed other limits such as CPU utilisation)
             and to set to Boolean.FALSE if we drop below our
             maxBandwidth * lowerThreshold; ie we apply
             hysteresis.

             Additionally we set this attribute name + ".overlt"
             to Boolean.TRUE if over maxBandwidth * lowerThreshold
             and to Boolean.FALSE otherwise (ie not applying
             hysteresis).  This can be used as a simple way
             of, for example, enabling compression but not
             changing the appearance of the site if not
             fully loaded but getting there.
             Note that ONLY bandwidth is used in this measure,
             though CPU and other factors may be used in the more
             general measure.
         &lt;/description&gt;
     &lt;/init-param&gt;
     &lt;init-param&gt;
         &lt;param-name&gt;throughputAttrNameBase&lt;/param-name&gt;
         &lt;param-value&gt;org.hd.d.pg2k.throughput&lt;/param-value&gt;
         &lt;description&gt;
             This creates several attributes at application
             scope with the given name plus a suffix.
         &lt;/description&gt;
     &lt;/init-param&gt;
 &lt;/filter&gt;

...

 &lt;!-- Apply throughput filter to all outputs and input.
      This should as close to the network stream as reasonably possible,
      eg *after* any compression filters in the outgoing stream,
      thus this needs to be an early filter-mapping in
      the web.xml file so that it puts on the nearly-outermost wrapper.
   --&gt;
 &lt;filter-mapping&gt;
     &lt;filter-name&gt;Throughput Monitor Filter&lt;/filter-name&gt;
     &lt;url-pattern&gt;/*&lt;/url-pattern&gt;
 &lt;/filter-mapping&gt;

</pre>
 * <p>
 * TODO: improve behaviour for "distributable" WAR servers.
 * This filter is not necessarily suitable for multi-JVM front-ends or where the
 *     WAR container otherwise creates multiple instances of this filter
 *     for what is logically one front-end.
 */
public class ThroughputMonitorFilterBase implements Filter
    {
    /**Default attribute name prefix for values posted to servlet context(s) and local value Map. */
    public static final String DEFAULT_ATTR_PREFIX = "org.hd.d.tmf";

    /**Attribute name suffix for (Integer) server-lifetime estimate of daily unique visitors. */
    public static final String ATTR_NAME_SUFFIX_DAILY_UNIQUE_VISITORS = "dailyUniqueVisitors";

    /**Attribute name suffix for (smoothed) fraction of ticks in which server was busy (non-idle). */
    public static final String ATTR_NAME_SUFFIX_BUSY_FRACTION_SMOOTHED = "nonIdleTicks";

    /**Attribute name suffix for long-term (smoothed) concurrent-connection load. */
    public static final String ATTR_NAME_SUFFIX_CONNECTION_LOAD_LONG_TERM = "ltConnLoad";

    /**Attribute name suffix for (smoothed) concurrent-connection load. */
    public static final String ATTR_NAME_SUFFIX_CONNECTION_LOAD = "connLoad";

    /**Attribute name suffix for bytes-per-second outbound over the server's lifetime. */
    public static final String ATTR_NAME_SUFFIX_SERVER_LIFETIME_BYTES_PER_SEC_OUTBOUND = "totalBpsOutbound";

    /**Attribute name suffix for bytes-per-second outbound. */
    public static final String ATTR_NAME_SUFFIX_BYTES_PER_SEC_OUTBOUND = "Bps";

    /**Name of servlet request-scope attribute used to note a"new" client.
     * Value posted is:
     * <ul>
     * <li>Boolean.FALSE for a non-sticky client
     * <li>Boolean.TRUE for a sticky client
     * </ul>
     */
    public static final String REQ_ATTR_NAME_NEW_CLIENT_FLAG = DEFAULT_ATTR_PREFIX + '.' + "newStickyClient";


    /**If true, only monitor GET operations, ie typical user traffic.
     * Ignoring data from (say) POST operations may help to ignore
     * slave/master bandwidth consumption for example.
     */
    public static final boolean MONITOR_GET_TRAFFIC_ONLY = false;

    /**The filter configuration object that we are associated with.
     * If this value is null, this filter instance is not currently configured.
     */
    private FilterConfig config;

    /**Allow derived classes to read the config info. */
    protected FilterConfig getConfig() { return(config); }

    /**Config parameter; name of flag to record overload, or null. */
    private String overloadFlagName;
    /**Config parameter; upper hysteresis level on overload flag, or null. */
    private String upperThreshold;
    /**Config parameter; lower hysteresis level on overload flag, or null. */
    private String lowerThreshold;
    /**Config parameter; max bandwidth in bytes/second before overload, or null. */
    private String maxBandwidth;
    /**Config parameter; system property name for max bandwidth in bytes/second before overload, or null. */
    private String maxBandwidthPropertyName;
    /**Config parameter; attr of prefix to write current throughput figures, or null. */
    private String throughputAttrNameBase;
    /**Config parameter; seconds before old client IP is forgotten, or null */
    private Integer clientIPPurgeTime;

    /**Default upper hysteresis value. */
    public static final float DEFAULT_UPPER_THRESHOLD = 0.9f;
    /**Default lower hysteresis value. */
    public static final float DEFAULT_LOWER_THRESHOLD = 0.6f;
    /**Default maximum bytes per second for overload; about a 1Mbps line minus overheads. */
    public static final int DEFAULT_MAX_BYTES_PER_SEC = 100000;
    /**Default upper-threshold bytes per second for overload. */
    private static final int DEFAULT_UPPERTHR_BYTES_PER_SEC = (int) (DEFAULT_MAX_BYTES_PER_SEC * DEFAULT_UPPER_THRESHOLD);
    /**Default lower-threshold bytes per second for overload. */
    private static final int DEFAULT_LOWERTHR_BYTES_PER_SEC = (int) (DEFAULT_MAX_BYTES_PER_SEC * DEFAULT_LOWER_THRESHOLD);
    /**Default client-IP memory time (seconds), positive; a power of two may be especially efficient.
     * Experience suggests that 128s makes the system a little too volatile,
     * including such things as the reaction time to the system becoming busy
     * which us derived from this value.
     */
    public static final int DEFAULT_CLIENT_IP_PURGE_TIME_S = 256;

    /**Max bytes per second for overload; set to a sensible default. */
    private int maxBytesPerSecOverload = DEFAULT_UPPERTHR_BYTES_PER_SEC;
    /**Min bytes per second for overload; set to a sensible default. */
    private int minBytesPerSecOverload = DEFAULT_LOWERTHR_BYTES_PER_SEC;

    /**Estimated overhead in bytes of a new HTTP/TCP/IP connection. */
    public static final int NEW_HTTP_OVERHEAD_BYTES_EST = 512;

    /**Our regular timer for status monitoring.
     * Only to be accessed under the instance lock
     * by destroy() and _startTasks().
     */
    private Timer timer;

    /**Target interval in milliseconds between runs of poll().
     * Something of the order of a second is probably right,
     * in terms of responsiveness to network traffic and
     * not consuming too much CPU.
     */
    public static final int TIMER_INTERVAL_MS = 1000;
    /**How long do we let the system `warm up' before the first tick?
     * Something of the order of a few seconds is probably right,
     * since the system may be pretty busy with other things to start with.
     */
    private static final int WARMUP_TIME_MS = 3 * TIMER_INTERVAL_MS;
    /**Time constant for smoothed throughput value, in ticks (calls to poll()).
     * Should be just enough to smooth out the effects of typical single
     * transfers, but small enough to allow actual overloads of
     * a few seconds duration (a reasonable human response time)
     * to be seen.
     * <p>
     * May be good for this to be similar to the client purge time.
     * <p>
     * A reasonable value is probably in the range 4--64,
     * with powers of two probably making for slightly faster computations.
     */
    private static final int SMOOTHER_TC_TICKS = Math.max(32,
        (DEFAULT_CLIENT_IP_PURGE_TIME_S * 1000) / TIMER_INTERVAL_MS);
    /**Time constant for long-term smoothed throughput value. */
    private static final int LTSMOOTHER_TC_TICKS = 1200;
    /**If true, take the max of medium and long-term throughput metrics to determine overload.
     * Taking the max may let the mechanism be quick to react to
     * short-term traffic bursts and improve the hysteresis where
     * there is a longer-term overload.
     */
    private static final boolean INCLUDE_LT_IN_OVERLOAD_SWITCH = true;

    /**Start time of this instance. */
    private final long filterInstanceStartTime = System.currentTimeMillis();
    /**Total outbound bytes through this instance.
     * Updated by poll() under the instance lock.
     * Must be read under the instance lock.
     */
    private long totalBytesOutbound; // Initially 0.
    /**Instantaneous bandwidth (bytes per second); updated by poll under the instance lock; private to poll(). */
    private int _bytesPerSecond;
    /**Smoothed bandwidth (bytes per second); updated by poll under the instance lock; private to poll(). */
    private int _bytesPerSecondSmoothed;
    /**Long-term smoothed bandwidth (bytes per second); updated by poll under the instance lock; private to poll(). */
    private double _ltBytesPerSecondSmoothed;
    /**Last poll time; private to poll(); starts with time instance was created.
     * Updated by poll() under the instance lock.
     */
    private long _lastPollTime = filterInstanceStartTime;

    /**Approximate measurement interval for gathering unqiue visitors, in milliseconds.
     * This is the size of the bucket in which uniqueness is measured.
     * If very small then users will be counted again if they continue to
     * use (or revisit) the site for (after) a short time;
     * if very large then the memory consumption could be excessive.
     * <p>
     * A period of a few hours to a few days is probably best,
     * since typical visits are from a few minutes to an hour or so,
     * and we should not see so many visitors in a few hours that we
     * strain memory resources.
     * <p>
     * (We also forcefully end the current measurement period and
     * start another when the number of entries in the current bucket
     * gets past some large threshold to put an upper bound on memory use.)
     */
    public static final int UNIQUE_VISTOR_MEASUREMENT_INTERVAL_MS =
        6 * 3600 * 1000; // Six hours; longer than typical visits.
    /**Limit on bucket size while counting unique visitors.
     * This acts as a cap on memory usage,
     * and by the time we've collected this many users we probably
     * have a decent sample size anyway!
     * <p>
     * Not expected to be invoked in practice unless the system gets
     * <em>very</em> busy.
     * <p>
     * A value of tens of thousands to a million or so would probably
     * limit memory whilst not significantly affecting accuracy,
     * even for a very busy site.
     */
    public static final int UNIQUE_VISTOR_MEASUREMENT_MAX_BUCKET_SIZE =
        100000;

    /**Count of unique visitors to site since start; never negative.
     * Accessed under the instance lock by, and private to, poll().
     */
    private int totalUniqueVisitorsCount;
    /**Set of IP addresses of unique human visitors in current period.
     * We are really only interested in the size of this,
     * so we might, for example, store masked addresses to
     * avoid overcounting accesses through proxies that have several
     * outbound addresses.
     * <p>
     * The accumulation period of this is typically of the order of
     * hours to days, ie in which continuing use by a single client
     * does not indicate a second visit, just one long visit.
     * Thus, many entries in here are from recent but not current users.
     * <p>
     * Accessed under the instance lock by, and private to, poll().
     */
    private final Set/*<InetAddress>*/ currentUniqueStickyVisitors = new HashSet/*<InetAddress>*/();
    /**Last time currentUniqueStickyVisitors was cleared.
     * Accessed under the instance lock by, and private to, poll().
     */
    private long _cuV_resetTime = filterInstanceStartTime;
    /**Map from (sticky) client IP address (InetAddress) to Long time of last client access.
     * Is a Hashtable for thread-safety.
     * <p>
     * Size of table is taken to be approximately number of sticky clients
     * currently using the site.
     * <p>
     * Sticky clients are those that appear to be real humans browsing
     * from one page to another, not just (say) hotlinking to one image.
     */
    private final Hashtable/*<InetAddress,Long>*/ stickyClientIPToTime = new Hashtable/*<InetAddress,Long>*/();
    /**Map from (all) client IP address (InetAddress) to Long time of last client access.
     * Is a Hashtable for thread-safety.
     * <p>
     * Size of table is taken to be approximately number of clients
     * currently using the site.
     * <p>
     * These clients are those that appear to be real humans
     * but may or may not be "sticky".
     */
    private final Hashtable/*<InetAddress,Long>*/ allClientIPToTime = new Hashtable/*<InetAddress,Long>*/();
    /**Bytes accumulated since last poll.
     * Updated by moreBytes() and zeroed by poll() under the instance lock.
     */
    private int _bytesSent;
    /**Count of currently open connections not blocked for input/output.
     * Never negative, only writable by adjustBusyConnectionCount(),
     * and under the instance lock.
     * <p>
     * Read by poll(); but is volatile so need not have a lock grabbed to be read.
     */
    private volatile int _connCount;
    /**Smoothed connection count analagous to UNIX run-queue length; updated by poll under the instance lock; private to poll(). */
    private float _connCountSmoothed;
    /**Long-term smoothed connection count analagous to UNIX run-queue length; updated by poll under the instance lock; private to poll(). */
    private float _ltConnCountSmoothed;
    /**Set of non-I/O-bound HTTP service threads being monitored in the filter; never null.
     * Private to adjustBusyConnectionCount() and protected by its lock.
     * <p>
     * We use a HashSet for speed.
     */
    private final Set/*<Thread>*/ _threadsInFilter = new HashSet/*<Thread>*/(131);
    /**Set to true if the system was idle, ie _connCount was zero, during a sample period.
     * Only writable by adjustBusyConnectionCount() (to set it true)
     * and poll() (to set it at the boundary of a sample period),
     * and under the instance lock.
     * <p>
     * Read by poll(), under the instance lock.
     */
    private boolean _wasIdle;
    /**Smoothed "was idle" value; updated by poll under the instance lock; private to poll().
     * Zero means was not idle at all, 1 means was idle at least some time every tick.
     * <p>
     * A starting value of 0.5f should help the server "slow-start",
     * without making it instantly "busy" unless genuinely strugging.
     */
    private float _wasIdleSmoothed = 0.5f;


    /**Map of values from attribute name to Integer/Float/Boolean; never null.
     * The key names have the same suffix as the values posted into
     * the servlet context (at application level),
     * and the same values.
     * <p>
     * This is thread safe and read/write to this class and deriving classes,
     * but a read-only wrapped version is available to outside callers.
     */
    private final Map/*<String, Object>*/ valueMap = new Hashtable/*<String, Object>*/(31);

    /**Read-only wrapper on valueMap; never null. */
    private final Map/*<String, Object>*/ valueMapRO = Collections.unmodifiableMap(valueMap);

    /**Read a value from the value map; the key must not be null, the result may be. */
    public Object getValue(final String s) { return(valueMap.get(s)); }

    /**Get a read-only view of the entire (String->Object) valueMap; never null. */
    public Map/*<String, Object>*/ getValueMapRO() { return(valueMapRO); }


    /**Run timer task. */
    private void poll()
        {
        final long now = System.currentTimeMillis();
        // Time since filter (and, presumably, the whole server) started.
        final long timeSinceStart = now - filterInstanceStartTime;
        // Time since the last call of poll(); strictly non-negativce.
        // It will appear to be (very) large on the first tick.
        final int timeSinceLastPoll = (int) Math.max(0,
            Math.min(now - _lastPollTime, Integer.MAX_VALUE));
        // Approximate (rounded) number of "ticks" passed; should normally be 1.
        // Zero meant the tick came early, eg because of scheduling funnies.
        // A value of 2 or more means that this poll() was delayed.
        final int ticksSinceLastPoll =
            (timeSinceLastPoll + TIMER_INTERVAL_MS/2) / TIMER_INTERVAL_MS;
        final boolean thisPollIsLate = (ticksSinceLastPoll > 1);
        final boolean thisPollIsVeryLate = thisPollIsLate &&
            (ticksSinceLastPoll >= SMOOTHER_TC_TICKS);
        if(thisPollIsVeryLate) { System.err.println("WARNING: ThroughputMonitorFilter: ticks missed: " + (ticksSinceLastPoll-1)); }

        int bpss; // Bytes-per-second smoothed.
        double ltbpss; // Long-term bytes-per-second smoothed.

        float ccs; // Connection-count smoothed.
        float ltccs; // Long-term connection-count smoothed.

        float busyFractionSmoothed; // Busy smoothed (0.0f means idle, 1.0f is busy).

        int dailyUniqueVisitors; // Computed count of visitors.

        long tBOut; // Local copy of totalBytesOutbound.

        // Fetch/compute bandwidth and other measures under the instance lock.
        //
        // We need not set every one of these on each tick,
        // especially those that only meaningfully change slowly.
        //
        // Minimise work done with the lock held to maximise concurrency,
        // but all read/write of mutable instance variables must happen here.
        synchronized(this)
            {
            // BANDWIDTH CALCULATION.
            // Note total bytes sent.
            totalBytesOutbound += _bytesSent;
            // TODO: fix this calc to be safe beyond 2GByte/s.
            _bytesPerSecond = (timeSinceLastPoll > 0) ? (int) (((_bytesSent*1000L)+(timeSinceLastPoll/2)) / timeSinceLastPoll) : 0;
            bpss = _bytesPerSecondSmoothed =
                ((_bytesPerSecondSmoothed * (SMOOTHER_TC_TICKS-1)) + _bytesPerSecond) / SMOOTHER_TC_TICKS;
            ltbpss = _ltBytesPerSecondSmoothed =
                ((_ltBytesPerSecondSmoothed * (LTSMOOTHER_TC_TICKS-1)) + _bytesPerSecond) / LTSMOOTHER_TC_TICKS;
            _lastPollTime = now;
            _bytesSent = 0;

            // CONNECTION COUNT CALCULATION.
            ccs = _connCountSmoothed =
                ((_connCountSmoothed * (SMOOTHER_TC_TICKS-1)) + _connCount) / SMOOTHER_TC_TICKS;
            ltccs = _ltConnCountSmoothed =
                ((_ltConnCountSmoothed * (LTSMOOTHER_TC_TICKS-1)) + _connCount) / LTSMOOTHER_TC_TICKS;

            // IDLE-STATE WATCHER.
            // We regard the machine as idle if it was in connection terms AND
            // no ticks were missed.
            final boolean idle = _wasIdle && !thisPollIsLate;
            // We run this once for every tick that has passed,
            // ie normally just once but not at all if not even one tick passed,
            // and more than once if ticks were missed
            // to more accurately reflect the real idleness of the machine.
            //
            // For example, if the machine was garbage collecting for so long
            // that we missed several ticks, this loop should approximate
            // the effect on _wasIdleSmoothed having been updated for a busy
            // machine on each lost tick during that GC when we were busy.
            //
            // To avoid burning loads of CPU pointlessly when we have missed
            // lots of ticks, for example on the first tick or when the
            // whole server process has been sleeping or paged out,
            // simply mark the system as fully busy instead of looping.
            if(thisPollIsVeryLate)
                { _wasIdleSmoothed = 0; } // Mark as busy.
            else for(int i = ticksSinceLastPoll; --i >= 0; )
                {
                _wasIdleSmoothed =
                    ((_wasIdleSmoothed * (SMOOTHER_TC_TICKS-1)) + (idle?1:0)) / SMOOTHER_TC_TICKS;
                }
            busyFractionSmoothed = 1.0f - _wasIdleSmoothed;
            // We note this next tick as starting idle iff:
            //   * there are currently no connections open, and,
            //   * we had not missed any ticks.
            _wasIdle = (_connCount == 0) && !thisPollIsLate;

            // Compute effective number of unique daily visitors since start.
            // TODO: should round to a unit inversely related to sample period?
            dailyUniqueVisitors = (int)
                ((((totalUniqueVisitorsCount + currentUniqueStickyVisitors.size()) *
                    (24 * 3600 * 1000L)) + (timeSinceStart / 2))
                       / timeSinceStart);

            tBOut = totalBytesOutbound;
            }

//System.err.println("Throughput (bytes/s): " + bpss + " [instantaneous: "+bps+"]");
//System.err.println("Was-Idle metric: " + wis);

        final ServletContext ctxt = config.getServletContext();

        // Compute higher of short-term-smoothed and long-term bandwidth values.
        // For good short-term response with good memory of recent behaviour.
        final int combinedBpss = (int) Math.round(Math.max(bpss, ltbpss));


        _purgeDepartedVisitors();
        // Get post-purge vistor count/estimate.
        final int stickyVisitorCount = stickyClientIPToTime.size();
        final int allVisitorCount = allClientIPToTime.size();

        // Update and recompute total-unique-visitors value
        // under the instance lock.
        // To maximise concurrency, minimise work done with the lock held.
        _updateCurrentUniqueStickyVisitors();


        // Compute a rough available mean bandwidth available to
        // a new client that would not make us too busy.
        // We report this as zero if we are feeling put-upon and
        // don't want more load right now.
        // We may also report zero if we have no users and thus
        // think that we might be unreachable.
        // We scale down reported bandwidth if the system is busy so as
        // to account for other resource limits.
        // (Note that we use a sub-linear scaling-back of reported bandwidth
        // with increasing numbers of users to reflect
        // the assumed bursty nature of user traffic.)
        // We set our system-load and bandwidth cutouts at the high thresholds
        // so as to allow a well-regulated load not to force us
        // to stop advertising as a mirror.
//        final boolean suspiciouslyQuiet = (stickyVisitorCount < 1);
        final boolean tooManyVisitors = (stickyVisitorCount/8) > computeApproximateUserCapacity();
        final boolean connectionTooBusy = combinedBpss >= maxBytesPerSecOverload;
        final boolean systemTooBusy = busyFractionSmoothed >= DEFAULT_UPPER_THRESHOLD;
        final int availableBpsPerClientBasic =
                (/* suspiciouslyQuiet || */ tooManyVisitors || connectionTooBusy || systemTooBusy) ? 0 :
                  Math.max(0, Math.min(maxBytesPerSecOverload, Math.round(
                        Math.min((maxBytesPerSecOverload - (float)combinedBpss) * (1-busyFractionSmoothed),
                                 minBytesPerSecOverload / (float)Math.sqrt(1+stickyVisitorCount)) )));


        // Allow a deriving class to do some work.
        // Pass it all the most critical computed values.
        tick(availableBpsPerClientBasic,
             allVisitorCount,
             stickyVisitorCount,
             (float) ltbpss, // Full double precision is probably meaningless.
             dailyUniqueVisitors,
             busyFractionSmoothed,
             tooManyVisitors, connectionTooBusy, systemTooBusy);


        // If the user has set throughputAttrNameBase,
        // post updates on throughput (and client count, etc) regularly.
        //
        // We need not set every one of these on each tick,
        // especially those that only meaningfully change slowly.
        if(throughputAttrNameBase != null)
            { _setAttributes(ctxt, bpss, tBOut, timeSinceStart, ccs, ltccs, busyFractionSmoothed, dailyUniqueVisitors); }

        // If the user has named an overload flag, set it appropriately
        // according to the appropriate smoothed bandwidth-consumption value
        // and measured CPU load and user count.
        // This should help improve user experience with available resources.
        if(overloadFlagName != null)
            { _setOverloadFlag(tooManyVisitors, ctxt, combinedBpss, bpss, busyFractionSmoothed); }


        // If the system seems to be too busy
        // (ie not servicing clients as fast as they can accept data)
        // then we might do such things as sample parts of the code-base active,
        // because of CPU load or anything other than bandwidth shortage.
        // Unless running on a multi-CPU system this will likely only catch
        // I/O-bound threads.
        if(systemTooBusy)
            { logBusyThreads(_connCount); }

        // Else if we detect that the system is completely idle
        // then we may trigger some special "idle" processing.
        //
        // Idle means:
        //   * we haven't just missed a tick (or more)
        //   * no current visitors
        //   * no live connections at this moment
        //   * no significant CPU/bandwidth load now/recently
        else if((!thisPollIsLate) &&
           (allVisitorCount == 0) &&
           (_connCount == 0) &&
           (busyFractionSmoothed <= (1-DEFAULT_UPPER_THRESHOLD)))
            { onIdle(); }
        }


    /**Set servlet attributes (and Map values) with various running parameters.
     * These parameters can then be picked up by other parts of the server,
     * eg for display in an Ops-only JSP.
     * <p>
     * Must only be called if throughputAttrNameBase is not null.
     */
    private void _setAttributes(final ServletContext ctxt,
                                final int bpss,
                                final long tBOut,
                                final long timeSinceStart,
                                final float ccs,
                                final float ltccs,
                                final float busyFractionSmoothed,
                                final int dailyUniqueVisitors)
        {
        _postValue(ctxt, throughputAttrNameBase,
            ATTR_NAME_SUFFIX_BYTES_PER_SEC_OUTBOUND,
            new Integer(bpss));
        _postValue(ctxt, throughputAttrNameBase,
            ATTR_NAME_SUFFIX_SERVER_LIFETIME_BYTES_PER_SEC_OUTBOUND,
            new Float((1000.0 * tBOut) / timeSinceStart));
        _postValue(ctxt, throughputAttrNameBase,
            ATTR_NAME_SUFFIX_CONNECTION_LOAD,
            new Float(ccs));
        _postValue(ctxt, throughputAttrNameBase,
            ATTR_NAME_SUFFIX_CONNECTION_LOAD_LONG_TERM,
            new Float(ltccs));
        _postValue(ctxt, throughputAttrNameBase,
            ATTR_NAME_SUFFIX_BUSY_FRACTION_SMOOTHED,
            new Float(busyFractionSmoothed));
        _postValue(ctxt, throughputAttrNameBase,
            ATTR_NAME_SUFFIX_DAILY_UNIQUE_VISITORS,
            new Integer(dailyUniqueVisitors));

//            ctxt.setAttribute(throughputAttrNameBase + ".maxBps",
//                new Integer(maxBytesPerSecOverload));
//        ctxt.setAttribute(throughputAttrNameBase + ".Bps",
//            new Integer(bpss));
//            ctxt.setAttribute(throughputAttrNameBase + ".ltBps",
//                new Double(ltbpss));
//            ctxt.setAttribute(throughputAttrNameBase + ".clientCount",
//                new Integer(stickyVisitorCount));
//        ctxt.setAttribute(throughputAttrNameBase + ".totalBpsOutbound",
//            new Float((1000.0 * tBOut ) / timeSinceStart));
//        ctxt.setAttribute(throughputAttrNameBase + ".connLoad",
//            new Float(ccs));
//        ctxt.setAttribute(throughputAttrNameBase + ".ltConnLoad",
//            new Float(ltccs));
//        ctxt.setAttribute(throughputAttrNameBase + ".nonIdleTicks",
//            new Float(busyFractionSmoothed));
//        ctxt.setAttribute(throughputAttrNameBase + '.' + ATTR_NAME_SUFFIX_DAILY_UNIQUE_VISITORS,
//            new Integer(dailyUniqueVisitors));
        }

    /**Post named value to everywhere that it needs to be recorded.
     * This is usually the servlet application-level context
     * and the valueMap.
     *
     * @param ctxt  servlet context to post to;
     *     never null unless throughputAttrNameBase is
     * @param throughputAttrNameBase  servlet contest attribute name prefix;
     *     never empty, and never null unless ctxt is
     *
     * @param name  the attribute name; not null, empty,
     *     not including any prefix nor leading '.'
     * @param value  the (immutable) attribute value;
     *     may be null to remove a value for the attribute
     */
    private void _postValue(final ServletContext ctxt,
                            final String throughputAttrNameBase,
                            final String name, final Object value)
        {
        assert((name != null) && (name.length() != 0) && !name.startsWith("."));

        // Post to the valueMap; note the special handling for null.
        if(null == value) { valueMap.remove(name); }
        else              { valueMap.put(name, value); }

        if(throughputAttrNameBase != null)
            { ctxt.setAttribute(throughputAttrNameBase + '.' + name, value); }
        }


    /**Called on each tick where this system seems to be idle.
     * Idle means:
     * <ul>
     * <li>we haven't just missed a tick (or more)
     * <li>no current visitors
     * <li>no live connections at this moment
     * <li>no significant CPU/bandwidth load now or recently
     * </ul>
     * <p>
     * Can be overridden in deriving class.
     */
    protected void onIdle()
        {
        // Do nothing in the base implementation.
        }

    /**Make a note of the thread state in a (busy) system.
     * Log the "busy" methods to the distributed logging store,
     * and extra detail to the System.err.
     * <p>
     * Can be overridden in deriving class.
     *
     * @param connCount  current active-connection count; non-negative
     */
    protected void logBusyThreads(final int connCount)
        {
        // Does nothing by default.
        }

    /**Called on every tick, and passed the most important metrics.
     * Can be overridden in deriving class.
     *
     * @param availableBpsPerClientBasic  estimated available bytes/sec
     *     available for one new client (or an existing client); non-negative
     * @param allVisitorCount  approximate count of "live" visitors;
     *     non-negative
     * @param stickyVisitorCount  approximate count of "live" "sticky" visitors;
     *     non-negative
     * @param ltbpss  long-term bytes/sec smoothed throughput
     * @param dailyUniqueVisitors  estimated unique vistors per day;
     *     non-negative
     * @param busyFractionSmoothed  smoothed system "busy-ness" value
     *     ranging from 0.0 (idle) to 1.0 (busy)
     * @param tooManyVisitors  true if well over capacity by user count
     * @param connectionTooBusy  true if well over capacity by bandwidth
     * @param systemTooBusy  true if system busy fraction dangerously high
     */
    protected void tick(final int availableBpsPerClientBasic,
                        final int allVisitorCount,
                        final int stickyVisitorCount,
                        final float ltbpss,
                        final int dailyUniqueVisitors,
                        final float busyFractionSmoothed,
                        final boolean tooManyVisitors,
                        final boolean connectionTooBusy,
                        final boolean systemTooBusy)
        {
        // Does nothing by default.
        }

    /**Purge entries for visitors that have not been active recently.
     * This covers "sticky" and "non-sticky" visitors.
     */
    private void _purgeDepartedVisitors()
        {
        final long now = System.currentTimeMillis();
        // Go through and eliminate any old clients that have gone away.
        // We work on keys from a copy of the tables to avoid blocking too long.
        //
        // Also compute a count of "sticky" visitors that have survived
        // at least a couple of purges.

        final long purgeTimeMs = 1000L * ((clientIPPurgeTime != null) ?
            clientIPPurgeTime.intValue() :
            DEFAULT_CLIENT_IP_PURGE_TIME_S);
        final HashSet/*<InetAddress>*/ keys1;
        synchronized(stickyClientIPToTime) { keys1 = new HashSet/*<InetAddress>*/(stickyClientIPToTime.keySet()); /* HashSet for speed. */ }
        for(final Iterator/*<InetAddress>*/ it = keys1.iterator(); it.hasNext(); )
            {
            synchronized(stickyClientIPToTime)
                {
                final InetAddress ia = (InetAddress) it.next();
                final Object o = stickyClientIPToTime.get(ia);
                if(!(o instanceof Long)) { continue; } // Key has gone away...
                final Long l = (Long) o;
                final long lastAccessTime = l.longValue();
                if(lastAccessTime + purgeTimeMs < now)
                    {
                    // Last visit too long ago: purge from clients and keys.
                    stickyClientIPToTime.remove(ia);
                    it.remove();
                    continue;
                    }
                }
            }
        keys1.clear(); // Help GC.

        final HashSet/*<InetAddress>*/ keys2;
        synchronized(allClientIPToTime) { keys2 = new HashSet/*<InetAddress>*/(allClientIPToTime.keySet()); /* HashSet for speed. */ }
        for(final Iterator/*<InetAddress>*/ it = keys2.iterator(); it.hasNext(); )
            {
            synchronized(allClientIPToTime)
                {
                final InetAddress ia = (InetAddress) it.next();
                final Object o = allClientIPToTime.get(ia);
                if(!(o instanceof Long)) { continue; } // Key has gone away...
                final Long l = (Long) o;
                final long lastAccessTime = l.longValue();
                if(lastAccessTime + purgeTimeMs < now)
                    {
                    // Last visit too long ago: purge from clients and keys.
                    allClientIPToTime.remove(ia);
                    it.remove();
                    continue;
                    }
                }
            }
        keys2.clear(); // Help GC.
        }

    /**Update the count of unique (daily) visitors.
     */
    private void _updateCurrentUniqueStickyVisitors()
        {
        final long now = System.currentTimeMillis();
        final HashSet/*<InetAddress>*/ keys = new HashSet/*<InetAddress>*/(stickyClientIPToTime.keySet());
        synchronized(this)
            {
            // Add current set of live sticky clients to unique totals...
            currentUniqueStickyVisitors.addAll(keys);

            // Need to update the count of unique visitors..
            //
            // Note that we omit the current count of users from the
            // amount added into the total to avoid double-counting users
            // across the measurement bucket boundary
            // (and we put the current users back into the new bucket),
            // to avoid causing a spike up or down in the total visitor count.
            if((now - _cuV_resetTime >= UNIQUE_VISTOR_MEASUREMENT_INTERVAL_MS) ||
               (currentUniqueStickyVisitors.size() >= UNIQUE_VISTOR_MEASUREMENT_MAX_BUCKET_SIZE))
                {
                final int newVistorsThisBucket =
                    currentUniqueStickyVisitors.size() - keys.size();
                assert(newVistorsThisBucket >= 0);
//System.err.println("[New visitors this bucket ("+(UNIQUE_VISTOR_MEASUREMENT_INTERVAL_MS/1000)+"s) = "+newVistorsThisBucket+".]");
                totalUniqueVisitorsCount += newVistorsThisBucket;
                currentUniqueStickyVisitors.clear();
                currentUniqueStickyVisitors.addAll(keys);
                _cuV_resetTime = now;
                }
            }
        keys.clear(); // Help GC.
        }

    /**If need be, set servlet-wide "overload" flag for rest of server to switch to cheaper behaviour.
     */
    private void _setOverloadFlag(final boolean tooManyVisitors,
                                  final ServletContext ctxt,
                                  final int combinedBpss,
                                  final int bpss,
                                  final float busyFractionSmoothed)
        {
        // If seriously overloaded by user count,
        // eg /.-ed or DDoSed or whatever,
        // go into forced overload mode.
        // We take "seriously" to be about an order of magnitude
        // more than expected user capacity regardless of bandwidth.
        if(tooManyVisitors)
            {
//System.err.println("OVERLOAD: USER COUNT: users=" + currentVisitors + ", capacity=" + capacity + ", maxBps=" + maxBytesPerSecOverload);
            ctxt.setAttribute(overloadFlagName, Boolean.TRUE);
            ctxt.setAttribute(overloadFlagName + ".overlt", Boolean.TRUE);
            }
        // Else, if in more sane territory of user numbers,
        // check for bandwidth and/or CPU overload.
        else
            {
            // Get current bandwidth.
            final int bwiu = INCLUDE_LT_IN_OVERLOAD_SWITCH ? combinedBpss : bpss;

            // Set value with hysteresis (from bandwidth and CPU figures).
            if((bwiu > maxBytesPerSecOverload) || (busyFractionSmoothed > DEFAULT_UPPER_THRESHOLD))
                { ctxt.setAttribute(overloadFlagName, Boolean.TRUE); }
            else if((bwiu < minBytesPerSecOverload) && (busyFractionSmoothed < DEFAULT_LOWER_THRESHOLD))
                { ctxt.setAttribute(overloadFlagName, Boolean.FALSE); }

            // Set a bandwidth measure of overload that does not use hysteresis.
            // This is good to drive (say) selective compression of HTML pages.
            ctxt.setAttribute(overloadFlagName + ".overlt",
                ((bwiu > minBytesPerSecOverload) ?
                 Boolean.TRUE : Boolean.FALSE ));
            }

//System.err.println("Attr " + overloadFlagName + " = " + ctxt.getAttribute(overloadFlagName));
//System.err.println("Attr " + overloadFlagName + " = " + ctxt.getAttribute(overloadFlagName + ".overlt"));
        }

    /**Computes approximate number if simultaneous users that this server instance can support; strictly positive.
     * The computed number of users should be comfortable to sustain
     * with normal traffic patterns,
     * but is at best an estimate.
     * <p>
     * Callers of this should expect its value to be a factor of 10-or-so out
     * (ie 10x too high or 10x too low).
     * <p>
     * Can be overridden.
     */
    protected int computeApproximateUserCapacity()
        {
        return(Math.max(1, (maxBytesPerSecOverload /
                        expectedBpsPerStickyClient())));
        }

    /**Returns expected approximate bytes-per-second per sticky client; strictly positive.
     * Defaults to a reasonable value that may or may not suit your application.
     * <p>
     * Can be overridden.
     * <p>
     * TODO: make this a parameter that can be set in web.xml.
     */
    protected int expectedBpsPerStickyClient()
        {
        return(4096); // Assume a few kB/sec.
        }


    /**Called by the output stream each time more bytes are passed on.
     * Updates bytesSent under the instance lock.
     * <p>
     * Package visible...
     */
    synchronized void moreBytes(int bytes)
        { if(bytes > 0) { _bytesSent += bytes; } }

    /**Adjust count of currently-open, busy connections (not blocked for I/O).
     * This attempts to count the number of connections blocked other than
     * for I/O, eg when CPU-bound or waiting for other non-bandwidth resources.
     * <p>
     * In effect maintains a count (1 or 0) per thread, rather than globally.
     * <p>
     * Uses of this can be safely nested for any one thread,
     * so long as the return value is noted,
     * and no attempt to perform the matching action is made if false.
     * <p>
     * Package visible...
     */
    final synchronized boolean adjustBusyConnectionCount(final boolean adjustUpwards)
        {
        final Thread t = Thread.currentThread();
        if(adjustUpwards)
            {
            // Already "live" for this thread...
            if(!_threadsInFilter.add(t)) { return(false); }
            // Going live for this thread....
            ++_connCount;
            }
        else
            {
            // Already "non-live" for this thread...
            if(!_threadsInFilter.remove(t)) { return(false); }
            // Going non-live for this thread, and possibly globally!
            if(--_connCount == 0) { _wasIdle = true; }

            assert(_connCount >= 0);
            }

        // Check all threads still live.
//        if(IsDebug.isDebug)
//            {
//            for(final Thread tt : _threadsInFilter)
//                {
//                switch(tt.getState())
//                    {
//                    case TERMINATED:
//                        {
//                        System.err.println("ERROR: ThroughputMonitorFilter: dead thread in _threadsInFilter removed: " + tt);
//                        _threadsInFilter.remove(tt);
//                        --_connCount;
//                        break;
//                        }
//                    }
//                }
//            }

        assert(_connCount == _threadsInFilter.size());

//if(ORG.hd.d.IsDebug.isDebug) { System.out.println("***** LIVE " + (adjustUpwards ? "^^^^^^^" : "vvvvvvv") + " connections: " + _connCount); }
        return(true); // Was able to do the requested adjustment...
        }

    /**Place this filter into service.
     * Start the background timer as soon as we have finished configuration.
     *
     * @param filterConfig  the filter configuration object
     */
    public void init(final FilterConfig filterConfig)
        {
        config = filterConfig;

        // Set sensible defaults...
        maxBytesPerSecOverload = DEFAULT_MAX_BYTES_PER_SEC;
        minBytesPerSecOverload = DEFAULT_LOWERTHR_BYTES_PER_SEC;
        overloadFlagName = null;

        if(filterConfig != null)
            {
            maxBandwidth = filterConfig.getInitParameter("maxBandwidth");
            maxBandwidthPropertyName = filterConfig.getInitParameter("maxBandwidthPropertyName");
            upperThreshold = filterConfig.getInitParameter("upperThreshold");
            lowerThreshold = filterConfig.getInitParameter("lowerThreshold");
            overloadFlagName = filterConfig.getInitParameter("overloadFlagName");

            throughputAttrNameBase = filterConfig.getInitParameter("throughputAttrNameBase");
            if(throughputAttrNameBase == null) { throughputAttrNameBase = DEFAULT_ATTR_PREFIX; }
            System.out.println("[ThroughputMonitorFilter: application/map attribute prefix: "+throughputAttrNameBase+".]");

            clientIPPurgeTime = null;
            try { clientIPPurgeTime = new Integer(filterConfig.getInitParameter("clientIPPurgeTime")); }
            catch(final NumberFormatException e) { /* Ignore unparseable values. */ }

            float upperthr_hysteresis = DEFAULT_UPPER_THRESHOLD;
            if(upperThreshold != null)
                {
                upperthr_hysteresis = Math.min(0.99f, Math.max(0.01f, Float.parseFloat(upperThreshold)));
                }

            float lowerthr_hysteresis = DEFAULT_LOWER_THRESHOLD;
            if(lowerThreshold != null)
                {
                lowerthr_hysteresis = Math.min(upperthr_hysteresis, Math.max(0.01f, Float.parseFloat(lowerThreshold)));
                }

            // Use the default max bytes per second,
            // or, if available, the web.xml parameter value,
            // or, if available, the system property value.
            int maxBytesPerSecond = DEFAULT_MAX_BYTES_PER_SEC;
            if(maxBandwidth != null)
                {
                maxBytesPerSecond = Math.max(1, Integer.parseInt(maxBandwidth));
                }
            if(maxBandwidthPropertyName != null)
                {
                maxBytesPerSecond = Integer.getInteger(maxBandwidthPropertyName, maxBytesPerSecond).intValue();
                }

            maxBytesPerSecOverload = Math.round(upperthr_hysteresis * maxBytesPerSecond);
            minBytesPerSecOverload = Math.round(lowerthr_hysteresis * maxBytesPerSecond);

            System.out.println("[ThroughputMonitorFilter: upper and lower bandwidth thresholds (byte/sec): " + maxBytesPerSecOverload + ", " + minBytesPerSecOverload + ".]");
            System.out.println("[ThroughputMonitorFilter: approximate user capacity: " + computeApproximateUserCapacity() + ".]");

            // Assume that during our "warm-up" time
            // the WAR server will also be warming up and having to
            // compile JSPs, load classes, etc, and so will be very busy.
            // This should help the system to "soft-start" with less CPU-
            // (and other resource-) intensive algorithms in some cases.
            if(overloadFlagName != null)
                { config.getServletContext().setAttribute(overloadFlagName, Boolean.TRUE); }

            // Start the timer (if not already started)...
            // We schedule it to run at roughly regular intervals,
            // but it does not matter if they are not entirely regular,
            // especially if the system is busy.
            synchronized(this)
                {
                if(timer == null)
                    {
                    timer = new Timer(true); // Make this timer a daemon.
                    timer.schedule((new TimerTask(){
                        public final void run()
                            {
                            try { poll(); }
                            catch(final Throwable t) { t.printStackTrace(); }
                            }
                        }), WARMUP_TIME_MS, TIMER_INTERVAL_MS);
                    }
                }
            }
        }

    /**Take this filter out of service.
     * Kill our timer and free other resources.
     */
    public synchronized void destroy()
        {
        final Timer tmpTimer = timer;
        if(tmpTimer != null) { tmpTimer.cancel(); }

        // Help GC.
        timer = null;
        config = null;
        }

    /**Called to handle each (HTTP) request.
     * Hooks into the requests to measure throughput and other stats.
     * <p>
     * May also redirect a client if a non-preferred (or unrecognised)
     * hostname is used to contact the server.
     */
    public void doFilter(final ServletRequest request,
                         final ServletResponse response,
                         final FilterChain chain)
        throws IOException, ServletException
        {
        // We can monitor this
        // if it is an HTTP request
        // and, possibly, only if it is a GET request.
        if((response instanceof HttpServletResponse) &&
           (request instanceof HttpServletRequest) &&
            (!MONITOR_GET_TRAFFIC_ONLY ||
                "GET".equalsIgnoreCase(((HttpServletRequest) request).getMethod())))
            {
            final InetAddress clientIP =
                InetAddress.getByName(request.getRemoteAddr());

            final HttpServletRequest requestH = (HttpServletRequest) request;

            // Add on the presumed overhead for accepting a new HTTP connection...
            moreBytes(NEW_HTTP_OVERHEAD_BYTES_EST);


            // Capture changes to standard session attributes.
            // Should only do this on GET requests
            // to avoid interfering with things such as file uploads (POST).
            final boolean isGET = MONITOR_GET_TRAFFIC_ONLY ||
                            "GET".equalsIgnoreCase(requestH.getMethod());


            // Record the hit from the client.
            // These are all users of the server,
            // not just sticky (human) visitors.
            if(null == allClientIPToTime.put(clientIP, new Long(System.currentTimeMillis())))
                { if(null == requestH.getSession(false)) { newClient(requestH, false); } }

            // Record the client as a (sticky) "user" if appropriate.
            if(isStickyClient(isGET, requestH, clientIP))
                {
                // Note the client's IP address and the current time.
                if(null == stickyClientIPToTime.put(clientIP, new Long(System.currentTimeMillis())))
                    { newClient(requestH, true); }
                }


            // Note that we are busy handling this connection,
            // then call down the chain...
            // Monitor connections for interactiveness unless asked not to.
            final boolean didIncrement = !dontMonitorMe(requestH) && adjustBusyConnectionCount(true); // Connection started.
            try {
                // Wrap the response to monitor the output (and input)...
                final ThroughputMonitorRequestWrapper requestWrapper = new ThroughputMonitorRequestWrapper(this,
                                                                                                           requestH);
                final ThroughputMonitorResponseWrapper responseWrapper = new ThroughputMonitorResponseWrapper(this,
                                                   (HttpServletResponse) response);
                try {
                    try { chain.doFilter(requestWrapper, responseWrapper); }
                    finally { requestWrapper.finishRequest(); }
                    }
                finally { responseWrapper.finishResponse(); }
                }
            finally
                { if(didIncrement) { adjustBusyConnectionCount(false); } } // Request complete.
            }
        else
            {
            // Can't wrap it, so let it go without getting involved...
            chain.doFilter(request, response);
            }
        }

    /**Called just before we would start monitoring this HTTP call for interactivity.
     * If this HTTP operation may unavoidably block for a long time
     * then have this return true to exclude this op from the stats.
     * <p>
     * This does not stop the bandwidth being monitored.
     * <p>
     * The default is only to measure the interactivity of GET operations,
     * since other ones (eg POST) are often not of interest
     * or involve unavoidably slow, non-idempotent operations.
     * <p>
     * Can be overridden in a deriving class.
     *
     * @param requestH  the request; never null
     * @return  true if this request should NOT be monitored for interactivity
     */
    protected boolean dontMonitorMe(final HttpServletRequest requestH)
        {
        // Only monitor GET operations for interactivity by default.
        return(!"GET".equals(requestH.getMethod()));
        }

    /**Called if servicing a monitored HTTP request from what appears to be a new client.
     * This is called when a new client "session" appears to be in progress;
     * this is called <em>before</em> the call is passed up the filter chain.
     * This is not a formal HTTP session with a session ID,
     * but rather a period of continuous activity from
     * a single IP address or block.
     * <p>
     * This will ignore what appear to be new "non-sticky" client requests
     * if an HTTP session is in place on the grounds that this is probably
     * a single client changing IP addresses.
     * "Sticky" requests are <em>not</em> filtered out if an HTTP session
     * is already in place since such as session may often be put in place
     * by the first hit and so clients "becoming" sticky might often get missed.
     * <p>
     * This relies on the client IP address so may see clients
     * behind some proxies as more than one client,
     * and may see multiple clients behind busy proxies as a single client.
     * <p>
     * This may be called zero or more times during any particular HTTP request.
     * <p>
     * Typically this will get called once for a new client on its first access,
     * and then once more when it has established that the client is "sticky".
     * <p>
     * This can be overridden by a deriving class,
     * though an overriding method should almost always call this.
     * <p>
     * In many cases the setting of the request attribute will prove
     * to be enough for an application using this filter.
     *
     * @param request  the HTTP request; never null
     * @param isSticky  true if the client appears to be "sticky"
     */
    protected void newClient(final HttpServletRequest request,
                             final boolean isSticky)
        {
        request.setAttribute(REQ_ATTR_NAME_NEW_CLIENT_FLAG,
                             isSticky ? Boolean.TRUE : Boolean.FALSE);
        }

    /**Is this hit from a "sticky" client?
     * Default behaviour is to regard nothing as "sticky".
     * <p>
     * This can be overridden by a deriving class
     * to provide flexible interpretation of what is "sticky".
     *
     * @param isGET  true if the method is GET
     * @param requestH  the HTTP request; never null
     * @param clientIP  the client's IP address; never null
     *
     * @return true  iff this hit is to be treated as from a "sticky" client
     */
    protected boolean isStickyClient(final boolean isGET,
                                     final HttpServletRequest requestH,
                                     final InetAddress clientIP)
        {
        return(false);
        }
    }
